package jp.nyatla.nyar4psg;

import javax.media.opengl.GL;

import jp.nyatla.nyartoolkit.core.squaredetect.NyARSquare;
import jp.nyatla.nyartoolkit.core.transmat.NyARTransMatResult;
import jp.nyatla.nyartoolkit.core.types.NyARDoublePoint2d;
import jp.nyatla.nyartoolkit.core.types.NyARDoublePoint3d;
import processing.core.PApplet;
import processing.core.PMatrix3D;
import processing.core.PVector;
import processing.core.PGraphics3D;
import processing.opengl.PGraphicsOpenGL;

/**
 * 
 * Companion class to NyARMultiBoard that simplifies multiple marker tracking using
 * NyARToolkit for processing.  See the NyARMultiBoard documentation for more details.
 * 
 * @author Charl P. Botha http://cpbotha.net/
 *
 */
public class NyARMultiBoardMarker {
	/**
	 * The angle value in radian unit of "x,y,z" .
	 */
	public PVector angle;
	/**
	 * The translation value in radian unit of "x,y,z".
	 */
	public PVector trans;
	/**
	 * The position of 4 corner of marker.
	 */
	public int[][] pos2d;
	/**
	 * The transform matrix of detected marker, OpenGL conventions.
	 */
	public double[] transmat;
	
	/**
	 * PMatrix3D version of the OpenGL-style transmat, kept up to date along with transmat.
	 */
	public PMatrix3D transmatP3D;

	/**
	 * used to store the projection and modelview matrices during P3D mode
	 * beginTransform() and endTransform() calls.
	 */
	protected PMatrix3D stored_projection = null;
	protected PMatrix3D stored_modelview = null;
	
	/**
	 * Was this marker detected during the previous round?
	 */
	public boolean detected = false;
	
	/**
	 * The confidence value of detected marker.
	 */
	public double confidence=0.0;
	
	/**
	 * It is a number in which it continuously lost the marker.
	 * This value range is "0&lt;n&lt;lostDelay"
	 */
	public int lostCount = 0;
	
	
	/**
	 * Store instance of the board that holds us.
	 */
	protected NyARMultiBoard _board;
	
	public NyARMultiBoardMarker(NyARMultiBoard board)
	{
		this._board = board;
	
		this.angle=new PVector();
		this.trans=new PVector();
		this.pos2d=new int[4][2];
		this.transmat=new double[16];
		this.transmatP3D = new PMatrix3D();
	}
			
	/**
	 * This function sets corresponding transform matrix to the surface of the marker.
	 * The coordinate system of processing moves to the surface of the marker when this function is executed.
	 * Must return the coordinate system by using endTransform function at the end.
	 * 
	 * modified by charlbotha.com:
	 * - automatically adapts to P3D or OPENGL renderers
	 * - you don't need to pass the graphics context anymore
	 */
	public void beginTransform()
	{
		PApplet pa = this._board.get_papplet();
		
		if (pa.g.getClass().getName().equals(pa.P3D))
		{
			PGraphics3D p3d = (PGraphics3D)pa.g;
			
			if (this.stored_projection != null)
				pa.die("beginTransformP3D has already been called.", null);
			
			this.stored_projection = p3d.projection.get();
			this.stored_modelview = p3d.modelview.get();
			
			p3d.projection.set(this._board.get_ps_projection());
			p3d.modelview.set(this.transmatP3D);		
		}
		else if (pa.g.getClass().getName().equals(pa.OPENGL))
		{	
			PGraphicsOpenGL i_pgl = (PGraphicsOpenGL)pa.g;
			if(this._gl!=null){
				pa.die("The function beginTransform is already called.", null);			
			}
			this._pgl=i_pgl;
			this._gl=this._pgl.beginGL();
	
			
			this._gl=i_pgl.gl;
			this._gl.glMatrixMode(GL.GL_PROJECTION);
			pa.pushMatrix();
			pa.resetMatrix();
			//PGraphicsOpenGLã�®updateProjectionã�®ãƒ¢ãƒŽãƒžãƒ�ã‚’ã�—ã�¾ã�™ã€‚
			this._gl.glLoadMatrixd(this._board.projection,0);
			this._old_matrix=this._pgl.projection;
			this._pgl.projection=this._board.get_ps_projection();
	
	
			this._gl.glMatrixMode(GL.GL_MODELVIEW);
			pa.pushMatrix();
			pa.resetMatrix();
			this._gl.glLoadMatrixd(this.transmat,0);
					
			pa.pushMatrix();
		}
		else
		{
			pa.die("beginTransform() only supported for P3D and OPENGL renderers.", null);
		}
	}
	
	/**
	 * This function recovers the coordinate system that was changed by beginTransform().
	 * 
	 * modified by charlbotha.com:
	 * - supports both P3D and OPENGL renderers
	 * - also see documentation of beginTransform()
	 */
	public void endTransform()
	{
		PApplet pa = this._board.get_papplet();

		if (pa.g.getClass().getName().equals(pa.P3D))
		{
			PGraphics3D p3d = (PGraphics3D)pa.g;
			
			if (this.stored_projection == null)
				pa.die("endTransformP3D() called without beginTransform3D().", null);
			
			p3d.projection.set(this.stored_projection);
			this.stored_projection = null;
			p3d.modelview.set(this.stored_modelview);
			this.stored_modelview = null;
		}		
		else if (pa.g.getClass().getName().equals(pa.OPENGL))
		{			
			if(this._gl==null){
				pa.die("The function beginTransform is never called.", null);			
			}
			this._pgl.projection=this._old_matrix;
			pa.popMatrix();
			pa.popMatrix();
			this._gl.glMatrixMode(GL.GL_PROJECTION);
			pa.popMatrix();
			this._gl.glMatrixMode(GL.GL_MODELVIEW);
			if(this._pgl!=null){
				this._pgl.endGL();
			}
			this._gl=null;
			this._pgl=null;
		}
		else
		{
			pa.die("endTransform() only supported for P3D and OPENGL renderers.", null);
		}		
	}	
	
	protected void updateTransmat(NyARSquare i_square,NyARTransMatResult i_src)
	{
		NyARPsgBaseClass.matResult2GLArray(i_src,this.transmat);
		
		// now convert OpenGL-style transmat to PMatrix3D
        transmatP3D.m00 = (float) transmat[0 + 0 * 4];
        transmatP3D.m01 = (float) transmat[0 + 1 * 4];
        transmatP3D.m02 = (float) transmat[0 + 2 * 4];
        transmatP3D.m03 = (float) transmat[0 + 3 * 4];
        transmatP3D.m10 = - (float) transmat[1 + 0 * 4];
        transmatP3D.m11 = - (float) transmat[1 + 1 * 4];
        transmatP3D.m12 = - (float) transmat[1 + 2 * 4];
        transmatP3D.m13 = - (float) transmat[1 + 3 * 4];
        transmatP3D.m20 = (float) transmat[2 + 0 * 4];
        transmatP3D.m21 = (float) transmat[2 + 1 * 4];
        transmatP3D.m22 = (float) transmat[2 + 2 * 4];
        transmatP3D.m23 = (float) transmat[2 + 3 * 4];
        transmatP3D.m30 = (float) transmat[3 + 0 * 4];
        transmatP3D.m31 = (float) transmat[3 + 1 * 4];
        transmatP3D.m32 = (float) transmat[3 + 2 * 4];
        transmatP3D.m33 = (float) transmat[3 + 3 * 4];
        
		
		
		//angle
		i_src.getZXYAngle(this._tmp_d3p);
		
		this.angle.x=(float)this._tmp_d3p.x;
		this.angle.y=(float)this._tmp_d3p.y;
		this.angle.z=(float)this._tmp_d3p.z;
		//trans
		this.trans.x=(float)i_src.m03;
		this.trans.y=(float)i_src.m13;
		this.trans.z=(float)i_src.m23;

		//poså��æ˜ 
		final NyARDoublePoint2d[] pts=i_square.sqvertex;
		for(int i=0;i<4;i++){
			this.pos2d[i][0]=(int)pts[i].x;
			this.pos2d[i][1]=(int)pts[i].y;
		}		
		
		return;	
	}
	/********
	 * 	protected/private
	 *******/
	private final NyARDoublePoint3d _tmp_d3p=new NyARDoublePoint3d();
	
	
	//ã‚­ãƒ£ãƒƒã‚·ãƒ¥ã�Ÿã�¡
	private GL _gl=null;
	private PGraphicsOpenGL _pgl=null;	
	private PMatrix3D _old_matrix;
	

}
